/*
 * Copyright (c) 2015-2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Jun  2 2015
 */
#include <errno.h>
#include <inttypes.h>
#include <signal.h>
#include <string.h>

#include "mex.h"

#include "libgenomix.h"


/* --- local data ---------------------------------------------------------- */

#define json_skip_whitespace(json) (json) += strspn(json, " \t\n\r")
#define JSON_MINBUF	128

static const char tok_null[] = "null";
static const char tok_true[] = "true";
static const char tok_false[] = "false";

static mxArray *	j2s_value(const char **json, char *emsg,
				size_t emsgsize);
static char *		s2j_input(const mxArray *input, char *emsg,
				size_t emsgsize);
static int		s2j_value(const char *key, const mxArray *value,
				const mxArray *format, char **json,
				char **jend, size_t *jlen, char *emsg,
				size_t emsgsize);

static int		bufcat(char **buf, char **end, size_t *len,
				int quoted, const char *data);


/* --- genomix_server ------------------------------------------------------ */

void __attribute__ ((visibility ("default")))
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  char emsg[256];
  char key[128];
  int s;

  if (nrhs < 1)
    mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

  mxGetString(prhs[0], key, sizeof(key));
  if (!strcmp(key, "get")) {
    char uri[1024];
    const char *r;
    int id;

    if (nrhs < 3)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

    id = mxGetScalar(prhs[1]);
    mxGetString(prhs[2], uri, sizeof(uri));

    if (nrhs > 3) {
      if (!mxIsStruct(prhs[3]))
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
      {
        int n = mxGetNumberOfFields(prhs[3]);
        const char *args[2*n+1];
        int argsfree[n];
        mxArray *v;
        int i, err;

        err = 0;
        for(i = 0; i < n; i++) {
          args[2*i] = mxGetFieldNameByNumber(prhs[3], i);
          v = mxGetFieldByNumber(prhs[3], 0, i);

          args[2*i+1] = NULL;
          argsfree[i] = 0;
          if (err) continue;

          if (mxIsStruct(v)) {
            /* convert struct to json */
            char *json = s2j_input(v, emsg, sizeof(emsg));
            if (!json) { err = 1; continue; }

            args[2*i+1] = json;
            argsfree[i] = 1;

          } else if (mxIsChar(v)) {
            /* assume properly encoded data */
            args[2*i+1] = mxArrayToString(v);
            if (!args[2*i+1]) { err = 1; continue; }
            argsfree[i] = 2;
          }
        }
        args[2*n] = NULL;

        r = err ? NULL : genomix_get(id, uri, args, emsg, sizeof(emsg));

        for(i = 0; i < n; i++) {
          if (!args[2*i+1]) continue;
          switch (argsfree[i]) {
            case 2: mxFree((char *)args[2*i+1]); break;
            case 1: free((char *)args[2*i+1]); break;
          }
        }
      }
    } else
      r = genomix_get(id, uri, NULL, emsg, sizeof(emsg));

    if (!r)
      mexErrMsgIdAndTxt("genomix:run", emsg);

    if (plhs > 0) {
      if (r[0] == '{' || r[0] == '[' || r[0] == '"')
        plhs[0] = j2s_value(&r, emsg, sizeof(emsg));
      else
        plhs[0] = mxCreateString(r);
      if (!plhs[0]) mexErrMsgIdAndTxt("genomix:run", emsg);
    }

  } else if (!strcmp(key, "doevents")) {
    int timeout = 0;

    if (nrhs > 1) {
      double *t;
      if (!mxIsDouble(prhs[1]))
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
      t = mxGetPr(prhs[1]);
      if (t) timeout = *t * 1000;
    }
    genomix_doevent(timeout);

  } else if (!strcmp(key, "json")) {
    const char *in, *json;

    if (nrhs != 2)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    in = mxArrayToString(prhs[1]);
    if (!in)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");

    json = in;
    if (nlhs > 0) {
      plhs[0] = j2s_value(&json, emsg, sizeof(emsg));
      if (!plhs[0]) mexErrMsgIdAndTxt("genomix:run", emsg);
    }
    mxFree((void *)in);


  } else if (!strcmp(key, "connect")) {
    char *host = NULL;
    char *session = NULL;

    if (nrhs > 3)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    if (nrhs > 2) {
      session = mxArrayToString(prhs[2]);
      if (!session)
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    }
    if (nrhs > 1) {
      host = mxArrayToString(prhs[1]);
      if (!host)
        mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    }

    s = genomix_connect(host, session, emsg, sizeof(emsg));

    if (host) mxFree(host);
    if (session) mxFree(session);
    if (s < 0) mexErrMsgIdAndTxt("genomix:run", emsg);

    if (!mexIsLocked()) mexLock();
    if (nlhs > 0)
      plhs[0] = mxCreateDoubleScalar(s);

  } else if (!strcmp(key, "disconnect")) {
    int id;

    if (nrhs != 2)
      mexErrMsgIdAndTxt("genomix:args", "invalid arguments");
    id = mxGetScalar(prhs[1]);

    if (!genomix_disconnect(id))
      mexUnlock();

  } else
    mexErrMsgIdAndTxt("genomix:args", "invalid command");
}


/* --- genomix_event ------------------------------------------------------- */

void
genomix_event(const char *data)
{
  mxArray *prhs[1];
  char emsg[256];

  prhs[0] = j2s_value(&data, emsg, sizeof(emsg));
  mexCallMATLAB(0, NULL, 1, prhs, "genomix.client.doevents");
  mxDestroyArray(prhs[0]);
}


/* --- j2s_tokerror -------------------------------------------------------- */

/** Append an error message with the current token
 */
static void *
j2s_tokerror(const char *what, const char *json, char *emsg, size_t emsgsize)
{
  char tok[128];
  size_t s;

  s = strcspn(json, " \t\n\r{}[],:\"");
  if (!s) s = 1;
  if (s > sizeof(tok)-1) s = sizeof(tok)-1;
  strncpy(tok, json, s);

  snprintf(emsg, emsgsize, "unexpected %s \"%s\"", what, tok);
  return NULL;
}


/* --- j2s_value ----------------------------------------------------------- */

/** Parse a JSON value
 */
static mxArray *
j2s_value(const char **json, char *emsg, size_t emsgsize)
{
  json_skip_whitespace(*json);
  switch(**json) {
    case 0:
      return NULL;

    case '{': /* object */ {
      mxArray **va = NULL;
      mxArray *k, *v, *s;
      char *key;
      size_t na = 0;
      int i, m, isarray;

      ++(*json);
      json_skip_whitespace(*json);

      s = mxCreateStructArray(1, &(mwSize){1}, 0, NULL);
      if (!s) return s;
      m = isarray = 0;

      while(**json != '}') {
        /* key: string */
        if (**json != '"')
          return j2s_tokerror("non-string key", *json, emsg, emsgsize);
        if (!(k = j2s_value(json, emsg, emsgsize))) return NULL;
        key = mxArrayToString(k);
        if (!key) return NULL;

        if (isarray || (key[0] >= '0' && key[0] <= '9')) {
          isarray = 1;
          i = atoi(key);
          if (i > m) m = i;
        } else
          i = mxAddField(s, key);
        mxFree(key);

        json_skip_whitespace(*json);
        if (**json != ':')
          return j2s_tokerror("element separator", *json, emsg, emsgsize);
        ++(*json);

        /* value */
        if (isarray) {
          if (i >= na) {
            size_t ona = na;
            if (na) na <<= 1; else na = 16;
            va = realloc(va, na * sizeof(*va));
            if (!va) return NULL;
            memset(&va[ona], 0, (char *)&va[na] - (char *)&va[ona]);
          }
          if (!(va[i] = j2s_value(json, emsg, emsgsize))) return NULL;
        } else {
          if (!(v = j2s_value(json, emsg, emsgsize))) return NULL;
          mxSetFieldByNumber(s, 0, i, v);
        }

        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        } else if (**json != '}')
          return j2s_tokerror("element", *json, emsg, emsgsize);
      }
      ++(*json);

      if (isarray) {
        mxDestroyArray(s);
        m++;
        s = mxCreateCellArray(1, &(mwSize){m});
        if (s) while(m--) if (va[m]) mxSetCell(s, m, va[m]);
        free(va);
      }

      return s;
    }

    case '[': /* array */ {
      mxArray *a, **v;
      size_t i, n;

      ++(*json);
      json_skip_whitespace(*json);

      i = 0;
      n = 16;
      v = malloc(n * sizeof(*v));
      if (!v) return NULL;
      while(**json != ']') {
        if (i >= n) {
          n <<= 1;
          v = realloc(v, n * sizeof(*v));
          if (!v) return NULL;
        }
        if (!(v[i] = j2s_value(json, emsg, emsgsize))) return NULL;

        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        } else if (**json != ']')
          return j2s_tokerror("element", *json, emsg, emsgsize);

        i++;
      }
      ++(*json);

      a = mxCreateCellArray(1, &(mwSize){i});
      if (a) while(i--) mxSetCell(a, i, v[i]);
      free(v);
      return a;
    }

    case '"': /* string */ {
      const char *start = ++(*json);
      int quote = 0;
      char *new, *p;
      mxArray *v;

      while (**json != '"' && **json) {
        if (**json == '\\') {
          if (!*((*json)++)) break;
          if (**json == '"') quote = 1;
        }
        (*json)++;
      }
      if (!**json)
        return j2s_tokerror("end of string", start, emsg, emsgsize);
      (*json)++;

      new = malloc(*json - start);
      if (!new) return NULL;
      strncpy(new, start, *json - start - 1);
      new[*json - start - 1] = 0;

      if (quote) {
        p = new;
        while((p = strstr(p, "\\"))) {
          if (*p == '"' || *p == '\\')
            memmove(p, p+1, strlen(p)/*including final \0*/);
          p++;
        }
      }
      v = mxCreateString(new);
      free(new);
      return v;
    }

    case 't': /* bare word: true, false, null */
    case 'f':
    case 'n':
      if (!strncmp(*json, tok_true, sizeof(tok_true)-1)) {
        *json += sizeof(tok_true)-1;
        return mxCreateLogicalScalar(true);
      }
      if (!strncmp(*json, tok_false, sizeof(tok_false)-1)) {
        *json += sizeof(tok_false)-1;
        return mxCreateLogicalScalar(false);
      }
      if (!strncmp(*json, tok_null, sizeof(tok_null)-1)) {
        *json += sizeof(tok_null)-1;
        return mxCreateStructArray(1, &(mwSize){1}, 0, NULL);
      }

      return j2s_tokerror("value", *json, emsg, emsgsize);

    default: /* number or error */ {
      const char *start = *json;
      char *end;
      double v;

      v = strtod(start, &end);
      if (end == start) return j2s_tokerror("value", start, emsg, emsgsize);
      *json = end;

      return mxCreateDoubleScalar(v);
    }
  }

  snprintf(emsg, emsgsize, "internal error");
  return NULL;
}


/* --- s2j_value ----------------------------------------------------------- */

/** Build a JSON value
 */
static char *
s2j_input(const mxArray *input, char *emsg, size_t emsgsize)
{
  mxArray *data, *format, *k, *v;
  const char *field;
  char *json, *end;
  size_t len;
  int i, n, s;

  data = mxGetField(input, 0, "data");
  format = mxGetField(input, 0, "format");
  if (!data || !format) {
    snprintf(emsg, emsgsize, "internal error");
    return NULL;
  }

  len = JSON_MINBUF;
  json = end = malloc(len);
  if (!json) {
    snprintf(emsg, emsgsize, "no memory");
    return NULL;
  }

  if ((s = bufcat(&json, &end, &len, 0, "{"))) goto error;
  n = mxGetNumberOfFields(format);
  for(i = 0; i < n; i++) {
    field = mxGetFieldNameByNumber(format, i);
    v = mxGetFieldByNumber(format, 0, i);
    k = mxGetField(data, 0, field);
    if (!k) {
      snprintf(emsg, emsgsize, "missing input value for argument `%s'", field);
      goto error;
    }

    if (i && bufcat(&json, &end, &len, 0, ",")) goto error;
    if (bufcat(&json, &end, &len, 0, "\"")) goto error;
    if (bufcat(&json, &end, &len, 0, field)) goto error;
    if (bufcat(&json, &end, &len, 0, "\":")) goto error;

    s = s2j_value(field, k, v, &json, &end, &len, emsg, emsgsize);
    if (s) goto error;
  }
  if (bufcat(&json, &end, &len, 0, "}")) goto error;
  return json;

error:
  free(json);
  return NULL;
}


static int
s2j_value(const char *key, const mxArray *value, const mxArray *format,
          char **json, char **jend, size_t *jlen, char *emsg, size_t emsgsize)
{
  char kind[64], skey[256];
  const char *field;
  mxArray *type, *stype, *svalue;
  mxArray *v;
  double *l;
  char *t;
  int i, n, s;

  v = mxGetField(format, 0, "kind");
  if (!v) goto fatal;
  mxGetString(v, kind, sizeof(kind));

  if (!strcmp(kind, "struct") || !strcmp(kind, "union")) {
    if (!mxIsStruct(value)) {
      snprintf(emsg, emsgsize, "`%s' input value must be a structure", key);
      return -1;
    }
    type = mxGetField(format, 0, "type");
    if (!type) goto fatal;

    /* check missing arguments */
    n = mxGetNumberOfFields(type);
    for(i = 0; i < n; i++) {
      field = mxGetFieldNameByNumber(type, i);
      v = mxGetField(value, 0, field);
      if (!v) {
        snprintf(emsg, emsgsize, "missing input argument `%s.%s'", key, field);
        return -1;
      }
    }

    /* build json */
    if ((s = bufcat(json, jend, jlen, 0, "{"))) return s;
    n = mxGetNumberOfFields(value);
    for(i = 0; i < n; i++) {
      field = mxGetFieldNameByNumber(value, i);
      svalue = mxGetFieldByNumber(value, 0, i);
      snprintf(skey, sizeof(skey), "%s.%s", key, field);
      stype = mxGetField(type, 0, field);
      if (!stype) {
        snprintf(emsg, emsgsize, "unknown input argument `%s'", skey);
        return -1;
      }

      if (i && (s = bufcat(json, jend, jlen, 0, ","))) return s;
      if ((s = bufcat(json, jend, jlen, 0, "\""))) return s;
      if ((s = bufcat(json, jend, jlen, 0, field))) return s;
      if ((s = bufcat(json, jend, jlen, 0, "\":"))) return s;

      s = s2j_value(skey, svalue, stype, json, jend, jlen, emsg, emsgsize);
      if (s) return s;
    }
    if ((s = bufcat(json, jend, jlen, 0, "}"))) return s;
    return 0;

  } else if (!strcmp(kind, "array") || !strcmp(kind, "sequence")) {
    if (!mxIsCell(value)) {
      snprintf(emsg, emsgsize, "%s %s expects cell array data", kind, key);
      return -1;
    }
    n = mxGetN(value);

    /* check missing arguments */
    type = mxGetField(format, 0, "type");
    if (!type) goto fatal;

    v = mxGetField(format, 0, "length");
    if (v) {
      l = mxGetPr(v);
      if (!l || n != (size_t)*l) {
        snprintf(emsg, emsgsize, "%s %s has wrong size: must be %zu elements",
                 kind, key, l?(size_t)*l:n);
        return -1;
      }
    }

    /* build json */
    if ((s = bufcat(json, jend, jlen, 0, "["))) return s;
    for(i = 0; i < n; i++) {
      svalue = mxGetCell(value, i);
      snprintf(skey, sizeof(skey), "%s[%d]", key, i);

      if (i && (s = bufcat(json, jend, jlen, 0, ","))) return s;

      s = s2j_value(skey, svalue, type, json, jend, jlen, emsg, emsgsize);
      if (s) return s;
    }
    if ((s = bufcat(json, jend, jlen, 0, "]"))) return s;

    return 0;

  } else if (!strcmp(kind, "optional")) {
    type = mxGetField(format, 0, "type");
    if (!type) goto fatal;

    if (mxIsEmpty(value)) {
      s = bufcat(json, jend, jlen, 0, tok_null);
    } else {
      s = s2j_value(key, value, type, json, jend, jlen, emsg, emsgsize);
    }
    return s;

  } else if (!strcmp(kind, "char") || !strcmp(kind, "string")) {
    if (!mxIsChar(value)) {
      snprintf(emsg, emsgsize, "%s %s expects string data", kind, key);
      return -1;
    }
    t = mxArrayToString(value);
    if (!t) goto fatal;
    n = strlen(t);

    v = mxGetField(format, 0, "length");
    if (v) {
      l = mxGetPr(v);
      if (!l || n != (size_t)*l) {
        snprintf(emsg, emsgsize, "%s %s too long: %zu max",
                 kind, key, l?(size_t)*l:n);
        return -1;
      }
    }

    if ((s = bufcat(json, jend, jlen, 0, "\""))) return s;
    if ((s = bufcat(json, jend, jlen, 1, t))) return s;
    if ((s = bufcat(json, jend, jlen, 0, "\""))) return s;

    mxFree(t);
    return 0;

  } else if (!strcmp(kind, "enum")) {
    if (!mxIsChar(value)) {
      snprintf(emsg, emsgsize, "%s %s expects string data", kind, key);
      return -1;
    }
    t = mxArrayToString(value);
    if (!t) goto fatal;

    v = mxGetField(format, 0, "values");
    if (!v) goto fatal;
    n = mxGetN(v);

    for(i = 0; i < n; i++) {
      svalue = mxGetCell(v, i);
      mxGetString(svalue, skey, sizeof(skey));
      if (!strncmp(skey, t, sizeof(skey))) {
        if ((s = bufcat(json, jend, jlen, 0, "\""))) return s;
        if ((s = bufcat(json, jend, jlen, 0, t))) return s;
        if ((s = bufcat(json, jend, jlen, 0, "\""))) return s;
        mxFree(t);
        return 0;
      }
    }

    snprintf(emsg, emsgsize, "bad value for %s %s: %s", kind, key, t);
    mxFree(t);
    return -1;

  } else {
    if (!mxIsNumeric(value)) {
      snprintf(emsg, emsgsize, "bad numeric value for %s %s", kind, key);
      return -1;
    }
    l = mxGetPr(value);
    if (!l) goto fatal;

    switch(mxGetClassID(value)) {
      case mxINT8_CLASS: case mxINT16_CLASS:
      case mxINT32_CLASS: case mxINT64_CLASS:
        snprintf(skey, sizeof(skey), "%" PRId64, (int64_t)*l);
        break;

      case mxUINT8_CLASS: case mxUINT16_CLASS:
      case mxUINT32_CLASS: case mxUINT64_CLASS:
        snprintf(skey, sizeof(skey), "%" PRIu64, (uint64_t)*l);
        break;

      case mxDOUBLE_CLASS: case mxSINGLE_CLASS:
        snprintf(skey, sizeof(skey), "%g", *l);
        break;

      default:
        snprintf(emsg, emsgsize, "bad numeric value for %s %s", kind, key);
        return -1;
    }
    if (!l) goto fatal;

    s = bufcat(json, jend, jlen, 0, skey);
    return s;
  }

fatal:
  snprintf(emsg, emsgsize, "internal error at '%s'", key);
  return -1;
}


/* --- bufcat -------------------------------------------------------------- */

static int
bufcat(char **buf, char **end, size_t *len, int quoted, const char *data)
{
  size_t avail;
  const char *s;
  char *d;

  s = data;
  while(*s) {
    avail = *len - (*end - *buf);
    if (quoted) {
      for(d = *end; *s && avail; avail--) {
        if (*s == '"' || *s == '\\') {
          if (avail < 2) {
            avail = 0;
            break;
          }
          *d++ = '\\';
          avail--;
        }
        *d++ = *s++;
      }
    } else {
      for(d = *end; *s && avail; avail--)
        *d++ = *s++;
    }

    if (!avail) {
      char *newbuf = (char *)realloc(*buf, *len << 1);
      if (!*newbuf) return ENOMEM;

      d = newbuf + (d - *buf);
      *buf = newbuf;
      *len <<= 1;
    }
    *d = 0;
    *end = d;
  }

  return 0;
}
